"""Platform config client."""

from importlib.metadata import version

from .client import ConfigClient, ConfigClientBase
from .entities import (
    AMDGPU,
    GPU,
    ACMEEnvironment,
    AMDGPUPreset,
    AppsConfig,
    ARecord,
    BucketsConfig,
    Cluster,
    DisksConfig,
    DNSConfig,
    EnergyConfig,
    EnergySchedule,
    EnergySchedulePeriod,
    GPUPreset,
    GrafanaConfig,
    IdleJobConfig,
    IngressConfig,
    IntelGPU,
    IntelGPUPreset,
    MonitoringConfig,
    NvidiaGPU,
    NvidiaGPUPreset,
    OrchestratorConfig,
    PatchClusterRequest,
    PatchOrchestratorConfigRequest,
    PrometheusConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)

__all__ = [
    "ConfigClient",
    "ConfigClientBase",
    "AMDGPU",
    "GPU",
    "ACMEEnvironment",
    "AddNodePoolRequest",
    "AMDGPUPreset",
    "AppsConfig",
    "ARecord",
    "BucketsConfig",
    "Cluster",
    "DisksConfig",
    "DNSConfig",
    "EnergyConfig",
    "EnergySchedule",
    "EnergySchedulePeriod",
    "GPUPreset",
    "GrafanaConfig",
    "IdleJobConfig",
    "IngressConfig",
    "IntelGPU",
    "IntelGPUPreset",
    "MonitoringConfig",
    "NvidiaGPU",
    "NvidiaGPUPreset",
    "OrchestratorConfig",
    "PatchClusterRequest",
    "PatchOrchestratorConfigRequest",
    "PrometheusConfig",
    "RegistryConfig",
    "ResourcePoolType",
    "ResourcePreset",
    "Resources",
    "SecretsConfig",
    "StorageConfig",
    "TPUPreset",
    "TPUResource",
    "VolumeConfig",
]
__version__ = version(__name__)
